"""
statsdmetrics
--------------
Metric classes for Statsd.

:license: released under the terms of the MIT license.
See LICENSE file for more information.
"""

from .metrics import (Counter, Timer, Gauge,
                      Set, GaugeDelta,
                      normalize_metric_name,
                      parse_metric_from_request,
                    )

__version__ = '0.3.0'

__all__ = (Counter, Timer, Gauge,
           Set, GaugeDelta,
           normalize_metric_name,
           parse_metric_from_request,
            )



"""
statsdmetrics.client.tcp
------------------------
Statsd clients to send metrics to server over TCP
"""

import socket

from . import (AutoClosingSharedSocket, AbstractClient,
        BatchClientMixIn, DEFAULT_PORT)


class TCPClientMixIn(object):
    """Mix-In class to send metrics over TCP"""

    def _create_socket(self):
        sock = AutoClosingSharedSocket(socket.socket(socket.AF_INET, socket.SOCK_STREAM))
        sock.add_client(self)
        sock.connect(self.remote_address)
        return sock

    def _request(self, data):
        self._socket.sendall("{}\n".format(data).encode())


class TCPClient(TCPClientMixIn, AbstractClient):
    """Statsd client using TCP to send metrics

    >>> client = TCPClient("stats.example.org")
    >>> client.increment("event")
    >>> client.increment("event", 3, 0.4)
    >>> client.decrement("event", rate=0.2)
    """

    def batch_client(self, size=512):
        """Return a TCP batch client with same settings of the TCP client"""

        batch_client = TCPBatchClient(self.host, self.port, self.prefix, size)
        self._configure_client(batch_client)
        return batch_client


class TCPBatchClient(BatchClientMixIn, TCPClientMixIn, AbstractClient):
    """Statsd client that buffers metrics and sends batch requests over TCP

    >>> client = TCPBatchClient("stats.example.org")
    >>> client.increment("event")
    >>> client.decrement("event.second", 3, 0.5)
    >>> client.flush()
    """

    def __init__(self, host, port=DEFAULT_PORT, prefix="", batch_size=512):
        AbstractClient.__init__(self, host, port, prefix)
        BatchClientMixIn.__init__(self, batch_size)

    def flush(self):
        """Send buffered metrics in batch requests over TCP"""

        while len(self._batches) > 0:
            self._socket.sendall(self._batches[0])
            self._batches.popleft()
        return self

    def unit_client(self):
        """Return a TCPClient with same settings of the batch TCP client"""

        client = TCPClient(self.host, self.port, self.prefix)
        self._configure_client(client)
        return client


__all__ = (TCPClient, TCPBatchClient)



"""
statsdmetrics.metrics
----------------------
Define metric classes
"""

from abc import ABCMeta, abstractmethod
from re import compile, sub

try:
    unicode('')
except NameError:
    unicode = str

try:
    long(1)
except NameError:
    long = int


def is_string(value):
    return isinstance(value, str) or \
           isinstance(value, unicode)


def is_numeric(value):
    return isinstance(value, int) or \
           isinstance(value, float) or \
           isinstance(value, long)


normalize_metric_name_regex_subs = (
    (compile("\s+"), "_"),
    (compile("[\/\\\\]"), "-"),
    (compile("[^\w.-]"), ""),
)


def normalize_metric_name(name):
    for regex, replacement in normalize_metric_name_regex_subs:
        name = sub(regex, replacement, name)
    return name


def parse_metric_from_request(request):
    assert is_string(request), \
        "Request should be string to parse a metric from"
    metric_type_classes = dict(
        c=Counter,
        ms=Timer,
        g=Gauge,
        s=Set,
    )
    metric_value_types = dict(
        c=int,
        ms=float,
        g=float,
    )

    name, data = request.split(':')
    value, _, type_section = data.partition('|')
    type_, _, sample_rate_section = type_section.partition('|@')

    if type_ not in metric_type_classes:
        raise ValueError(
            "Invalid request. Metric type '{}' is not supported".format(type_)
        )

    if type_ == 'g' and len(value) > 1 and value[0] in ('+', '-'):
        metric_class = GaugeDelta
    else:
        metric_class = metric_type_classes[type_]

    value = metric_value_types[type_](value) \
        if type_ in metric_value_types else value

    sample_rate = AbstractMetric.default_sample_rate \
        if sample_rate_section == '' else float(sample_rate_section)

    return metric_class(name.strip(), value, sample_rate)


class AbstractMetric(object):
    __metaclass__ = ABCMeta

    default_sample_rate = 1

    def __init__(self, name):
        self._name = ''
        self._sample_rate = self.__class__.default_sample_rate
        self.name = name

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name):
        assert is_string(name), \
            'Metric name should be string'
        assert name != '', \
            'Metric name should not be empty'
        self._name = name

    @property
    def sample_rate(self):
        return self._sample_rate

    @sample_rate.setter
    def sample_rate(self, value):
        assert is_numeric(value), \
            'Metric sample rate should be numeric: {}:{}'.format(
                self.name, value)
        assert value > 0, \
            'Metric sample rate should be positive: {}:{}'.format(
                self.name, value)
        self._sample_rate = value

    @abstractmethod
    def to_request(self):
        raise NotImplementedError()


class Counter(AbstractMetric):
    def __init__(self, name, count=0, sample_rate=1):
        super(Counter, self).__init__(name)
        self._count = 0
        self.count = count
        self.sample_rate = sample_rate

    @property
    def count(self):
        return self._count

    @count.setter
    def count(self, count):
        assert isinstance(count, int), \
            'Counter count should be integer'
        self._count = count

    def to_request(self):
        result = "{0}:{1}|c".format(self._name, self._count)
        if self._sample_rate != 1:
            result += "|@{:n}".format(self._sample_rate)
        return result

    def __eq__(self, other):
        assert isinstance(other, Counter), \
            'Counter can be compared to Counter only'
        return self.name == other.name \
               and self.count == other.count \
               and self.sample_rate == other.sample_rate

    def __ne__(self, other):
        assert isinstance(other, Counter), \
            'Counter can be compared to Counter only'
        return self.name != other.name \
               or self.count != other.count \
               or self.sample_rate != other.sample_rate


class Timer(AbstractMetric):
    def __init__(self, name, milliseconds, sample_rate=1):
        super(Timer, self).__init__(name)
        self._milliseconds = 0
        self.milliseconds = milliseconds
        self.sample_rate = sample_rate

    @property
    def milliseconds(self):
        return self._milliseconds

    @milliseconds.setter
    def milliseconds(self, milliseconds):
        assert is_numeric(milliseconds), \
            'Timer milliseconds should be numeric'
        assert milliseconds >= 0, \
            'Timer milliseconds should not be negative'
        self._milliseconds = milliseconds

    def to_request(self):
        result = "{0}:{1}|ms".format(self._name, self._milliseconds)
        if self._sample_rate != 1:
            result += "|@{:n}".format(self._sample_rate)
        return result

    def __eq__(self, other):
        assert isinstance(other, Timer), \
            'Timer can be compared to Timer only'
        return self.name == other.name \
               and self.milliseconds == other.milliseconds \
               and self.sample_rate == other.sample_rate

    def __ne__(self, other):
        assert isinstance(other, Timer), \
            'Timer can be compared to Timer only'
        return self.name != other.name \
               or self.milliseconds != other.milliseconds \
               or self.sample_rate != other.sample_rate


class Gauge(AbstractMetric):
    def __init__(self, name, value, sample_rate=1):
        self._value = 0
        super(Gauge, self).__init__(name)
        self.value = value
        self.sample_rate = sample_rate

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        assert is_numeric(value), \
            'Gauge value should be numeric'
        assert value >= 0, \
            'Gauge value should not be negative'
        self._value = value

    def to_request(self):
        result = "{0}:{1}|g".format(self._name, self._value)
        if self._sample_rate != 1:
            result += "|@{:n}".format(self._sample_rate)
        return result

    def __eq__(self, other):
        assert isinstance(other, Gauge), \
            'Gauge can be compared to Gauge only'
        return self.name == other.name \
               and self.value == other.value \
               and self.sample_rate == other.sample_rate

    def __ne__(self, other):
        assert isinstance(other, Gauge), \
            'Gauge can be compared to Gauge only'
        return self.name != other.name \
               or self.value != other.value \
               or self.sample_rate != other.sample_rate


class Set(AbstractMetric):
    def __init__(self, name, value, sample_rate=1):
        self._value = 0
        super(Set, self).__init__(name)
        self.value = value
        self.sample_rate = sample_rate

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        try:
            str(value)
        except (TypeError, ValueError):
            raise AssertionError(
                "Set value should be convertible to string")
        try:
            hash(value)
        except (TypeError, ValueError):
            raise AssertionError(
                "Set value should be hashable")
        self._value = value

    def to_request(self):
        result = "{0}:{1}|s".format(self._name, self._value)
        if self._sample_rate != 1:
            result += "|@{:n}".format(self._sample_rate)
        return result

    def __eq__(self, other):
        assert isinstance(other, Set), \
            "Set can be compared to Set only"
        return self.name == other.name \
               and self.value == other.value \
               and self.sample_rate == other.sample_rate

    def __ne__(self, other):
        assert isinstance(other, Set), \
            "Set can be compared to Set only"
        return self.name != other.name \
               or self.value != other.value \
               or self.sample_rate != other.sample_rate


class GaugeDelta(AbstractMetric):
    def __init__(self, name, delta, sample_rate=1):
        self._delta = 0
        super(GaugeDelta, self).__init__(name)
        self.delta = delta
        self.sample_rate = sample_rate

    @property
    def delta(self):
        return self._delta

    @delta.setter
    def delta(self, delta):
        assert is_numeric(delta), \
            "Gauge delta should be numeric"
        self._delta = delta

    def to_request(self):
        result = "{}:{:+n}|g".format(self._name, self._delta)
        if self._sample_rate != 1:
            result += "|@{:n}".format(self._sample_rate)
        return result

    def __eq__(self, other):
        assert isinstance(other, GaugeDelta), \
            'GaugeDelta can be compared to GaugeDelta only'
        return self.name == other.name \
               and self.delta == other.delta \
               and self.sample_rate == other.sample_rate

    def __ne__(self, other):
        assert isinstance(other, GaugeDelta), \
            'GaugeDelta can be compared to GaugeDelta only'
        return self.name != other.name \
               or self.delta != other.delta \
               or self.sample_rate != other.sample_rate


__all__ = (Counter, Timer, Gauge,
           Set, GaugeDelta,
           normalize_metric_name,
           parse_metric_from_request
           )



#!/usr/bin/env python

"""
statsdmetrics
--------------

Data metrics for Statsd.

"""
from __future__ import print_function

import os
from os.path import dirname

try:
    import setuptools
    from setuptools import setup
except ImportError:
    setuptools = None
    from distutils.core import setup

try:
    import distutilazy.test
    import distutilazy.clean
except ImportError:
    distutilazy = None

from statsdmetrics import __version__

classifiers = [
    "Development Status :: 4 - Beta",
    "Intended Audience :: Developers",
    "License :: OSI Approved :: MIT License",
    "Operating System :: OS Independent",
    "Programming Language :: Python",
    "Programming Language :: Python :: 2",
    "Programming Language :: Python :: 2.7",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.2",
    "Programming Language :: Python :: 3.3",
    "Programming Language :: Python :: 3.4",
    "Programming Language :: Python :: 3.5",
    "Programming Language :: Python :: Implementation :: CPython",
    "Programming Language :: Python :: Implementation :: PyPy",
    "Programming Language :: Python :: Implementation :: Jython",
    "Topic :: Software Development :: Libraries :: Python Modules",
    "Topic :: System :: Networking :: Monitoring"
]

long_description = __doc__
with open(os.path.join(os.path.dirname(__file__), "README.rst")) as fh:
    long_description = fh.read()

setup_params = dict(
    name = "statsdmetrics",
    packages = ["statsdmetrics", "statsdmetrics.client"],
    version = __version__,
    description = "Metric classes for Statsd",
    long_description = long_description,
    author = "Farzad Ghanei",
    author_email = "farzad.ghanei@gmail.com",
    url = "https://github.com/farzadghanei/statsd-metrics",
    license = "MIT",
    classifiers = classifiers,
)

if setuptools:
    dev_dependencies_filename = os.path.join(
        dirname(__file__),
        "requirements-dev.txt"
    )
    if os.path.exists(dev_dependencies_filename):
        with open(dev_dependencies_filename) as fh:
            setup_params["extras_require"] = {
                "dev": [item.strip() for item in fh.readlines()]
            }
    setup_params["keywords"] = "statsd metrics"
    setup_params["test_suite"] = "tests"
    setup_params["zip_safe"] = True

if distutilazy:
    setup_params["cmdclass"] = dict(
        test=distutilazy.test.run_tests,
        clean_pyc=distutilazy.clean.clean_pyc,
        clean=distutilazy.clean.clean_all
    )

if __name__ == "__main__":
    setup(**setup_params)

__all__ = (setup_params, classifiers, long_description)



"""
statsdmetrics.client
--------------------
Statsd client to send metrics to server
"""

import socket
from abc import ABCMeta
from random import random
from collections import deque

from ..metrics import (Counter, Timer, Gauge, GaugeDelta, Set,
                      normalize_metric_name, is_numeric)


DEFAULT_PORT = 8125


class AutoClosingSharedSocket(object):
    """Decorate sockets to attach metadata required by clients,
    and release system resources asap.

    The socket object is shared between multiple clients to use
    and will automatically close the socket when there are
    no more clients for the socket.
    """

    def __init__(self, sock):
        self._closed = False
        self._socket = sock
        self._clients = deque()

    @property
    def closed(self):
        return self._closed

    def close(self):
        """Close the socket to free system resources.

        After the socket is closed, further operations with socket
        will fail. Multiple calls to close will have no effect.
        """

        if self._closed:
            return
        self._socket.close()
        self._closed = True

    def add_client(self, client):
        """Add a client as a user of the socket.

        As long as the socket has users, it keeps the underlying
        socket object open for operations.
        """

        self._clients.append(id(client))

    def remove_client(self, client):
        """Remove the client from the users of the socket.

        If there are no more clients for the socket, it
        will close automatically.
        """

        try:
            self._clients.remove(id(client))
        except ValueError:
            pass

        if len(self._clients) < 1:
            self.close()

    def __del__(self):
        self.close()

    def __getattr__(self, name):
        return getattr(self._socket, name)


class AbstractClient(object):
    __metaclass__ = ABCMeta

    def __init__(self, host, port=DEFAULT_PORT, prefix=''):
        self._port = None
        self._host = host
        self._remote_address = None
        self._socket = None
        self.prefix = prefix
        self._set_port(port)
        self._socket = self._create_socket()

    @property
    def port(self):
        return self._port

    def _set_port(self, port):
        port = int(port)
        assert 0 < port < 65536
        self._port = port

    @property
    def host(self):
        return self._host

    @property
    def remote_address(self):
        if self._remote_address is None:
            self._remote_address = (socket.gethostbyname(self.host), self.port)
        return self._remote_address

    def increment(self, name, count=1, rate=1):
        """Increment a Counter metric"""

        if self._should_send_metric(name, rate):
            self._request(
                Counter(
                    self._create_metric_name_for_request(name),
                    int(count),
                    rate
                ).to_request()
            )

    def decrement(self, name, count=1, rate=1):
        """Decrement a Counter metric"""

        if self._should_send_metric(name, rate):
            self._request(
                Counter(
                    self._create_metric_name_for_request(name),
                    -1 * int(count),
                    rate
                ).to_request()
            )

    def timing(self, name, milliseconds, rate=1):
        """Send a Timer metric with the specified duration in milliseconds"""

        if self._should_send_metric(name, rate):
            if not is_numeric(milliseconds):
                milliseconds = float(milliseconds)
            self._request(
                Timer(
                    self._create_metric_name_for_request(name),
                    milliseconds,
                    rate
                ).to_request()
            )

    def gauge(self, name, value, rate=1):
        """Send a Gauge metric with the specified vlaue"""

        if self._should_send_metric(name, rate):
            if not is_numeric(value):
                value = float(value)
            self._request(
                Gauge(
                    self._create_metric_name_for_request(name),
                    value,
                    rate
                ).to_request()
            )

    def gauge_delta(self, name, delta, rate=1):
        """Send a GaugeDelta metric to change a Gauge by the specified value"""

        if self._should_send_metric(name, rate):
            if not is_numeric(delta):
                delta = float(delta)
            self._request(
                GaugeDelta(
                    self._create_metric_name_for_request(name),
                    delta,
                    rate
                ).to_request()
            )

    def set(self, name, value, rate=1):
        """Send a Set metric with the specified unique value"""

        if self._should_send_metric(name, rate):
            value = str(value)
            self._request(
                Set(
                    self._create_metric_name_for_request(name),
                    value,
                    rate
                ).to_request()
            )

    def _create_metric_name_for_request(self, name):
        return self.prefix + normalize_metric_name(name)

    def _should_send_metric(self, name, rate):
        return rate >= 1 or random() <= rate

    def _create_socket(self):
        sock = AutoClosingSharedSocket(socket.socket(socket.AF_INET, socket.SOCK_DGRAM))
        sock.add_client(self)
        return sock

    def _request(self, data):
        self._socket.sendto(str(data).encode(), self.remote_address)

    def _configure_client(self, other):
        other._remote_address = self._remote_address
        other._socket = self._socket
        self._socket.add_client(other)

    def __del__(self):
        if self._socket:
            self._socket.remove_client(self)
            self._socket = None


class BatchClientMixIn(object):
    """MixIn class to clients that buffer metrics and send batch requests"""

    def __init__(self, batch_size=512):
        batch_size = int(batch_size)
        assert batch_size > 0, "BatchClient batch size should be positive"
        self._batch_size = batch_size
        self._batches = deque()

    @property
    def batch_size(self):
        return self._batch_size

    def clear(self):
        """Clear buffered metrics"""

        self._batches.clear()
        return self

    def flush(self):
        """Send buffered metrics in batch requests"""

        address = self.remote_address
        while len(self._batches) > 0:
            self._socket.sendto(self._batches[0], address)
            self._batches.popleft()
        return self

    def _request(self, data):
        """Override parent by buffering the metric instead of sending now"""

        data = bytearray("{}\n".format(data).encode())
        self._prepare_batches_for_storage(len(data))
        self._batches[-1].extend(data)

    def _prepare_batches_for_storage(self, data_size=None):
        batch_size = self._batch_size
        data_size = data_size or batch_size
        if data_size > batch_size:
            self._batches.append(bytearray())
        elif len(self._batches) < 1 or\
                        (len(self._batches[-1]) + data_size) >= batch_size:
            self._batches.append(bytearray())

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.flush()


class Client(AbstractClient):
    """Statsd client, using UDP to send metrics

    >>> client = Client("stats.example.org")
    >>> client.increment("event")
    >>> client.increment("event", 3, 0.4)
    >>> client.decrement("event", rate=0.2)
    """

    def batch_client(self, size=512):
        """Return a batch client with same settings of the client"""

        batch_client = BatchClient(self.host, self.port, self.prefix, size)
        self._configure_client(batch_client)
        return batch_client


class BatchClient(BatchClientMixIn, AbstractClient):
    """Statsd client buffering requests and send in batch UDP requests

    >>> client = BatchClient("stats.example.org")
    >>> client.increment("event")
    >>> client.decrement("event.second", 3, 0.5)
    >>> client.flush()
    """

    def __init__(self, host, port=DEFAULT_PORT, prefix="", batch_size=512):
        AbstractClient.__init__(self, host, port, prefix)
        BatchClientMixIn.__init__(self, batch_size)

    def unit_client(self):
        """Return a client with same settings of the batch client"""

        client = Client(self.host, self.port, self.prefix)
        self._configure_client(client)
        return client

__all__ = (Client, BatchClient)
